# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p
from .MoneroAccountPublicAddress import MoneroAccountPublicAddress


class MoneroTransactionDestinationEntry(p.MessageType):
    FIELDS = {
        1: ('amount', p.UVarintType, 0),
        2: ('addr', MoneroAccountPublicAddress, 0),
        3: ('is_subaddress', p.BoolType, 0),
    }

    def __init__(
        self,
        amount: int = None,
        addr: MoneroAccountPublicAddress = None,
        is_subaddress: bool = None,
    ) -> None:
        self.amount = amount
        self.addr = addr
        self.is_subaddress = is_subaddress



class NetworkTypes(object):
    MAINNET = 0
    TESTNET = 1
    STAGENET = 2
    FAKECHAIN = 3


class MainNet(object):
    PUBLIC_ADDRESS_BASE58_PREFIX = 18
    PUBLIC_INTEGRATED_ADDRESS_BASE58_PREFIX = 19
    PUBLIC_SUBADDRESS_BASE58_PREFIX = 42


class TestNet(object):
    PUBLIC_ADDRESS_BASE58_PREFIX = 53
    PUBLIC_INTEGRATED_ADDRESS_BASE58_PREFIX = 54
    PUBLIC_SUBADDRESS_BASE58_PREFIX = 63


class StageNet(object):
    PUBLIC_ADDRESS_BASE58_PREFIX = 24
    PUBLIC_INTEGRATED_ADDRESS_BASE58_PREFIX = 25
    PUBLIC_SUBADDRESS_BASE58_PREFIX = 36


def net_version(network_type=NetworkTypes.MAINNET, is_subaddr=False):
    """
    Network version bytes used for address construction
    :return:
    """
    c_net = None
    if network_type is None or network_type == NetworkTypes.MAINNET:
        c_net = MainNet
    elif network_type == NetworkTypes.TESTNET:
        c_net = TestNet
    elif network_type == NetworkTypes.STAGENET:
        c_net = StageNet
    else:
        raise ValueError("Unknown network type: %s" % network_type)

    prefix = (
        c_net.PUBLIC_ADDRESS_BASE58_PREFIX
        if not is_subaddr
        else c_net.PUBLIC_SUBADDRESS_BASE58_PREFIX
    )
    return bytes([prefix])



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class ResetDevice(p.MessageType):
    MESSAGE_WIRE_TYPE = 14
    FIELDS = {
        1: ('display_random', p.BoolType, 0),
        2: ('strength', p.UVarintType, 0),  # default=256
        3: ('passphrase_protection', p.BoolType, 0),
        4: ('pin_protection', p.BoolType, 0),
        5: ('language', p.UnicodeType, 0),  # default=english
        6: ('label', p.UnicodeType, 0),
        7: ('u2f_counter', p.UVarintType, 0),
        8: ('skip_backup', p.BoolType, 0),
    }

    def __init__(
        self,
        display_random: bool = None,
        strength: int = None,
        passphrase_protection: bool = None,
        pin_protection: bool = None,
        language: str = None,
        label: str = None,
        u2f_counter: int = None,
        skip_backup: bool = None,
    ) -> None:
        self.display_random = display_random
        self.strength = strength
        self.passphrase_protection = passphrase_protection
        self.pin_protection = pin_protection
        self.language = language
        self.label = label
        self.u2f_counter = u2f_counter
        self.skip_backup = skip_backup



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class MoneroTransactionSetOutputRequest(p.MessageType):
    FIELDS = {
        1: ('dst_entr', p.BytesType, 0),
        2: ('dst_entr_hmac', p.BytesType, 0),
    }

    def __init__(
        self,
        dst_entr: bytes = None,
        dst_entr_hmac: bytes = None,
    ) -> None:
        self.dst_entr = dst_entr
        self.dst_entr_hmac = dst_entr_hmac



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class FirmwareUpload(p.MessageType):
    MESSAGE_WIRE_TYPE = 7
    FIELDS = {
        1: ('payload', p.BytesType, 0),  # required
        2: ('hash', p.BytesType, 0),
    }

    def __init__(
        self,
        payload: bytes = None,
        hash: bytes = None,
    ) -> None:
        self.payload = payload
        self.hash = hash



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: https://github.com/monero-project/mininero
# Author: Dusan Klinec, ph4r05, 2018

import gc

from apps.monero.xmr import crypto


def key_zero_vector(rows):
    """
    Empty key vector
    :param rows:
    :return:
    """
    vct = [None] * rows
    for i in range(rows):
        vct[i] = crypto.sc_0()
    return vct


def d2b(n, digits):
    b = [0] * digits
    i = 0
    while n:
        b[i] = n & 1
        i = i + 1
        n >>= 1
    return b


def prove_range_mem(amount, last_mask=None):
    """
    Memory optimized range proof.

    Gives C, and mask such that \sumCi = C
    c.f. http:#eprint.iacr.org/2015/1098 section 5.1

    Ci is a commitment to either 0 or 2^i, i=0,...,63
    thus this proves that "amount" is in [0, 2^ATOMS]
    mask is a such that C = aG + bH, and b = amount
    :param amount:
    :param last_mask: ai[ATOMS-1] will be computed as \sum_{i=0}^{ATOMS-2} a_i - last_mask
    :param use_asnl: use ASNL, used before Borromean
    :return: sumCi, mask, RangeSig.
        sumCi is Pedersen commitment on the amount value. sumCi = aG + amount*H
        mask is "a" from the Pedersent commitment above.
    """
    res = bytearray(32 * (64 + 64 + 64 + 1))
    mv = memoryview(res)
    gc.collect()

    def as0(mv, x, i):
        crypto.encodeint_into(x, mv[32 * i :])

    def as1(mv, x, i):
        crypto.encodeint_into(x, mv[32 * 64 + 32 * i :])

    def aci(mv, x, i):
        crypto.encodepoint_into(x, mv[32 * 64 * 2 + 32 + 32 * i :])

    n = 64
    bb = d2b(amount, n)  # gives binary form of bb in "digits" binary digits
    ai = key_zero_vector(n)
    a = crypto.sc_0()

    C = crypto.identity()
    alpha = key_zero_vector(n)
    c_H = crypto.gen_H()
    kck = crypto.get_keccak()  # ee computation

    # First pass, generates: ai, alpha, Ci, ee, s1
    for ii in range(n):
        ai[ii] = crypto.random_scalar()
        if last_mask is not None and ii == 64 - 1:
            ai[ii] = crypto.sc_sub(last_mask, a)

        a = crypto.sc_add(
            a, ai[ii]
        )  # creating the total mask since you have to pass this to receiver...

        alpha[ii] = crypto.random_scalar()
        L = crypto.scalarmult_base(alpha[ii])

        if bb[ii] == 0:
            Ctmp = crypto.scalarmult_base(ai[ii])
        else:
            Ctmp = crypto.point_add(crypto.scalarmult_base(ai[ii]), c_H)
        C = crypto.point_add(C, Ctmp)
        aci(mv, Ctmp, ii)

        if bb[ii] == 0:
            si = crypto.random_scalar()
            c = crypto.hash_to_scalar(crypto.encodepoint(L))
            L = crypto.add_keys2(si, c, crypto.point_sub(Ctmp, c_H))
            kck.update(crypto.encodepoint(L))
            as1(mv, si, ii)

        else:
            kck.update(crypto.encodepoint(L))

        c_H = crypto.point_double(c_H)

    # Compute ee, memory cleanup
    ee = crypto.sc_reduce32(crypto.decodeint(kck.digest()))
    crypto.encodeint_into(ee, mv[64 * 32 * 2 :])
    del kck
    gc.collect()

    # Second phase computes: s0, s1
    c_H = crypto.gen_H()

    for jj in range(n):
        if not bb[jj]:
            s0 = crypto.sc_mulsub(ai[jj], ee, alpha[jj])

        else:
            s0 = crypto.random_scalar()
            Ctmp = crypto.decodepoint(
                mv[32 * 64 * 2 + 32 + 32 * jj : 32 * 64 * 2 + 32 + 32 * jj + 32]
            )
            LL = crypto.add_keys2(s0, ee, Ctmp)
            cc = crypto.hash_to_scalar(crypto.encodepoint(LL))
            si = crypto.sc_mulsub(ai[jj], cc, alpha[jj])
            as1(mv, si, jj)
        as0(mv, s0, jj)

        c_H = crypto.point_double(c_H)

    gc.collect()
    return C, a, res


def prove_range(
    amount, last_mask=None, decode=False, backend_impl=True, byte_enc=True, rsig=None
):
    """
    Range proof generator.
    In order to minimize the memory consumption and CPU usage during transaction generation the returned values
    are returned encoded.

    :param amount:
    :param last_mask:
    :param backend_impl: backend implementation, if available
    :param decode: decodes output
    :param byte_enc: byte encoded
    :param rsig: buffer for rsig
    :return:
    """
    if not backend_impl or not byte_enc or decode:
        raise ValueError("Unsupported params")

    C, a, R = None, None, None
    try:
        if rsig is None:
            rsig = bytearray(32 * (64 + 64 + 64 + 1))

        buf_ai = bytearray(4 * 9 * 64)
        buf_alpha = bytearray(4 * 9 * 64)
        C, a, R = crypto.prove_range(
            rsig, amount, last_mask, buf_ai, buf_alpha
        )  # backend returns encoded

    finally:
        import gc

        buf_ai = None
        buf_alpha = None
        gc.collect()

    return C, a, R


# Ring-ct MG sigs
# Prove:
#   c.f. http:#eprint.iacr.org/2015/1098 section 4. definition 10.
#   This does the MG sig on the "dest" part of the given key matrix, and
#   the last row is the sum of input commitments from that column - sum output commitments
#   this shows that sum inputs = sum outputs
# Ver:
#   verifies the above sig is created corretly


def ecdh_encode(unmasked, receiver_pk=None, derivation=None):
    """
    Elliptic Curve Diffie-Helman: encodes and decodes the amount b and mask a
    where C= aG + bH
    :param unmasked:
    :param receiver_pk:
    :param derivation:
    :return:
    """
    from apps.monero.xmr.serialize_messages.tx_ecdh import EcdhTuple

    rv = EcdhTuple()
    if derivation is None:
        esk = crypto.random_scalar()
        rv.senderPk = crypto.scalarmult_base(esk)
        derivation = crypto.encodepoint(crypto.scalarmult(receiver_pk, esk))

    sharedSec1 = crypto.hash_to_scalar(derivation)
    sharedSec2 = crypto.hash_to_scalar(crypto.encodeint(sharedSec1))

    rv.mask = crypto.sc_add(unmasked.mask, sharedSec1)
    rv.amount = crypto.sc_add(unmasked.amount, sharedSec2)
    return rv


def ecdh_decode(masked, receiver_sk=None, derivation=None):
    """
    Elliptic Curve Diffie-Helman: encodes and decodes the amount b and mask a
    where C= aG + bH
    :param masked:
    :param receiver_sk:
    :param derivation:
    :return:
    """
    from apps.monero.xmr.serialize_messages.tx_ecdh import EcdhTuple

    rv = EcdhTuple()

    if derivation is None:
        derivation = crypto.scalarmult(masked.senderPk, receiver_sk)

    sharedSec1 = crypto.hash_to_scalar(derivation)
    sharedSec2 = crypto.hash_to_scalar(crypto.encodeint(sharedSec1))

    rv.mask = crypto.sc_sub(masked.mask, sharedSec1)
    rv.amount = crypto.sc_sub(masked.amount, sharedSec2)
    return rv


#
# Key image import / export
#


def generate_ring_signature(prefix_hash, image, pubs, sec, sec_idx, test=False):
    """
    Generates ring signature with key image.
    void crypto_ops::generate_ring_signature()

    :param prefix_hash:
    :param image:
    :param pubs:
    :param sec:
    :param sec_idx:
    :param test:
    :return:
    """
    from apps.monero.xmr.common import memcpy

    if test:
        from apps.monero.xmr import monero

        t = crypto.scalarmult_base(sec)
        if not crypto.point_eq(t, pubs[sec_idx]):
            raise ValueError("Invalid sec key")

        k_i = monero.generate_key_image(crypto.encodepoint(pubs[sec_idx]), sec)
        if not crypto.point_eq(k_i, image):
            raise ValueError("Key image invalid")
        for k in pubs:
            crypto.ge_frombytes_vartime_check(k)

    image_unp = crypto.ge_frombytes_vartime(image)
    image_pre = crypto.ge_dsm_precomp(image_unp)

    buff_off = len(prefix_hash)
    buff = bytearray(buff_off + 2 * 32 * len(pubs))
    memcpy(buff, 0, prefix_hash, 0, buff_off)
    mvbuff = memoryview(buff)

    sum = crypto.sc_0()
    k = crypto.sc_0()
    sig = []
    for i in range(len(pubs)):
        sig.append([crypto.sc_0(), crypto.sc_0()])  # c, r

    for i in range(len(pubs)):
        if i == sec_idx:
            k = crypto.random_scalar()
            tmp3 = crypto.scalarmult_base(k)
            crypto.encodepoint_into(tmp3, mvbuff[buff_off : buff_off + 32])
            buff_off += 32

            tmp3 = crypto.hash_to_ec(crypto.encodepoint(pubs[i]))
            tmp2 = crypto.scalarmult(tmp3, k)
            crypto.encodepoint_into(tmp2, mvbuff[buff_off : buff_off + 32])
            buff_off += 32

        else:
            sig[i] = [crypto.random_scalar(), crypto.random_scalar()]
            tmp3 = crypto.ge_frombytes_vartime(pubs[i])
            tmp2 = crypto.ge_double_scalarmult_base_vartime(sig[i][0], tmp3, sig[i][1])
            crypto.encodepoint_into(tmp2, mvbuff[buff_off : buff_off + 32])
            buff_off += 32

            tmp3 = crypto.hash_to_ec(crypto.encodepoint(tmp3))
            tmp2 = crypto.ge_double_scalarmult_precomp_vartime(
                sig[i][1], tmp3, sig[i][0], image_pre
            )
            crypto.encodepoint_into(tmp2, mvbuff[buff_off : buff_off + 32])
            buff_off += 32

            sum = crypto.sc_add(sum, sig[i][0])

    h = crypto.hash_to_scalar(buff)
    sig[sec_idx][0] = crypto.sc_sub(h, sum)
    sig[sec_idx][1] = crypto.sc_mulsub(sig[sec_idx][0], sec, k)
    return sig


def check_ring_singature(prefix_hash, image, pubs, sig):
    """
    Checks ring signature generated with generate_ring_signature
    :param prefix_hash:
    :param image:
    :param pubs:
    :param sig:
    :return:
    """
    from apps.monero.xmr.common import memcpy

    image_unp = crypto.ge_frombytes_vartime(image)
    image_pre = crypto.ge_dsm_precomp(image_unp)

    buff_off = len(prefix_hash)
    buff = bytearray(buff_off + 2 * 32 * len(pubs))
    memcpy(buff, 0, prefix_hash, 0, buff_off)
    mvbuff = memoryview(buff)

    sum = crypto.sc_0()
    for i in range(len(pubs)):
        if crypto.sc_check(sig[i][0]) != 0 or crypto.sc_check(sig[i][1]) != 0:
            return False

        tmp3 = crypto.ge_frombytes_vartime(pubs[i])
        tmp2 = crypto.ge_double_scalarmult_base_vartime(sig[i][0], tmp3, sig[i][1])
        crypto.encodepoint_into(tmp2, mvbuff[buff_off : buff_off + 32])
        buff_off += 32

        tmp3 = crypto.hash_to_ec(crypto.encodepoint(pubs[i]))
        tmp2 = crypto.ge_double_scalarmult_precomp_vartime(
            sig[i][1], tmp3, sig[i][0], image_pre
        )
        crypto.encodepoint_into(tmp2, mvbuff[buff_off : buff_off + 32])
        buff_off += 32

        sum = crypto.sc_add(sum, sig[i][0])

    h = crypto.hash_to_scalar(buff)
    h = crypto.sc_sub(h, sum)
    return crypto.sc_isnonzero(h) == 0


def export_key_image(
    creds,
    subaddresses,
    pkey,
    tx_pub_key,
    additional_tx_pub_keys,
    out_idx,
    test=True,
    verify=True,
):
    """
    Generates key image for the TXO + signature for the key image
    :param creds:
    :param subaddresses:
    :param pkey:
    :param tx_pub_key:
    :param additional_tx_pub_keys:
    :param out_idx:
    :param test:
    :param verify:
    :return:
    """
    from apps.monero.xmr import monero

    r = monero.generate_key_image_helper(
        creds, subaddresses, pkey, tx_pub_key, additional_tx_pub_keys, out_idx
    )
    xi, ki, recv_derivation = r[:3]

    phash = crypto.encodepoint(ki)
    sig = generate_ring_signature(phash, ki, [pkey], xi, 0, test)

    if verify:
        if check_ring_singature(phash, ki, [pkey], sig) != 1:
            raise ValueError("Signature error")

    return ki, sig



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import unittest

from monero_glue.xmr.core.backend import ed25519


class Basetest(unittest.TestCase):
    """Simple tests"""

    def __init__(self, *args, **kwargs):
        super(Basetest, self).__init__(*args, **kwargs)

    def test_ed25519(self):
        """
        Basic ED25519 invariants
        :return:
        """
        assert ed25519.b >= 10
        assert 8 * len(ed25519.H(b"hash input")) == 2 * ed25519.b
        assert pow(2, ed25519.q - 1, ed25519.q) == 1
        assert ed25519.q % 4 == 1
        assert pow(2, ed25519.l - 1, ed25519.l) == 1
        assert ed25519.l >= 2 ** (ed25519.b - 4)
        assert ed25519.l <= 2 ** (ed25519.b - 3)
        assert pow(ed25519.d, (ed25519.q - 1) // 2, ed25519.q) == ed25519.q - 1
        assert pow(ed25519.I, 2, ed25519.q) == ed25519.q - 1
        assert ed25519.isoncurve(ed25519.B)
        P = ed25519.scalarmult(ed25519.B, ed25519.l)
        assert ed25519.isoncurve(P)

    def test_ed25519_mult8(self):
        """
        Mult8 test vector
        :return:
        """
        p1 = ed25519.decodepoint(
            bytes(
                [
                    0x14,
                    0x9c,
                    0xe9,
                    0x3f,
                    0xd3,
                    0x1d,
                    0xf5,
                    0x44,
                    0x18,
                    0xa5,
                    0x29,
                    0x9b,
                    0x11,
                    0xdf,
                    0xc1,
                    0x4e,
                    0x48,
                    0xd1,
                    0x4a,
                    0x2a,
                    0x4c,
                    0xe9,
                    0x84,
                    0x0c,
                    0xf2,
                    0xff,
                    0x7c,
                    0x31,
                    0x83,
                    0xda,
                    0x89,
                    0x53,
                ]
            )
        )
        p2 = ed25519.decodepoint(
            bytes(
                [
                    0x25,
                    0x9e,
                    0xf2,
                    0xab,
                    0xa8,
                    0xfe,
                    0xb4,
                    0x73,
                    0xcf,
                    0x39,
                    0x05,
                    0x8a,
                    0x0f,
                    0xe3,
                    0x0b,
                    0x9f,
                    0xf6,
                    0xd2,
                    0x45,
                    0xb4,
                    0x2b,
                    0x68,
                    0x26,
                    0x68,
                    0x7e,
                    0xbd,
                    0x6b,
                    0x63,
                    0x12,
                    0x8a,
                    0xff,
                    0x64,
                ]
            )
        )
        self.assertEqual(ed25519.scalarmult(p1, 8), p2)

    def test_ed25519_scalarmult(self):
        """
        Scalar mult test vector
        :return:
        """
        pub = bytes(
            [
                0x22,
                0x4f,
                0xad,
                0x08,
                0x5c,
                0x7c,
                0x8b,
                0xe0,
                0xc0,
                0x58,
                0x48,
                0x46,
                0xf7,
                0x88,
                0xd9,
                0x48,
                0x1e,
                0x53,
                0x53,
                0xfb,
                0xa7,
                0xbb,
                0xff,
                0x4d,
                0x24,
                0xb0,
                0xb8,
                0xbc,
                0x43,
                0xbb,
                0x66,
                0x6f,
            ]
        )
        scal = ed25519.decodeint(
            bytes(
                [
                    0x4c,
                    0xe8,
                    0x8c,
                    0x16,
                    0x8e,
                    0x0f,
                    0x5f,
                    0x8d,
                    0x65,
                    0x24,
                    0xf7,
                    0x12,
                    0xd5,
                    0xf8,
                    0xd7,
                    0xd8,
                    0x32,
                    0x33,
                    0xb1,
                    0xe7,
                    0xa2,
                    0xa6,
                    0x0b,
                    0x5a,
                    0xba,
                    0x52,
                    0x06,
                    0xcc,
                    0x0e,
                    0xa2,
                    0xbc,
                    0x08,
                ]
            )
        )

        res = ed25519.scalarmult(ed25519.decodepointcheck(pub), scal)
        res_h = ed25519.encodepoint(res)

        exp_mul = bytes(
            [
                0x14,
                0x9c,
                0xe9,
                0x3f,
                0xd3,
                0x1d,
                0xf5,
                0x44,
                0x18,
                0xa5,
                0x29,
                0x9b,
                0x11,
                0xdf,
                0xc1,
                0x4e,
                0x48,
                0xd1,
                0x4a,
                0x2a,
                0x4c,
                0xe9,
                0x84,
                0x0c,
                0xf2,
                0xff,
                0x7c,
                0x31,
                0x83,
                0xda,
                0x89,
                0x53,
            ]
        )

        self.assertEqual(res_h, exp_mul)


if __name__ == "__main__":
    unittest.main()  # pragma: no cover



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p
from .LiskTransactionAsset import LiskTransactionAsset


class LiskTransactionCommon(p.MessageType):
    FIELDS = {
        1: ('type', p.UVarintType, 0),
        2: ('amount', p.UVarintType, 0),  # default=0
        3: ('fee', p.UVarintType, 0),
        4: ('recipient_id', p.UnicodeType, 0),
        5: ('sender_public_key', p.BytesType, 0),
        6: ('requester_public_key', p.BytesType, 0),
        7: ('signature', p.BytesType, 0),
        8: ('timestamp', p.UVarintType, 0),
        9: ('asset', LiskTransactionAsset, 0),
    }

    def __init__(
        self,
        type: int = None,
        amount: int = None,
        fee: int = None,
        recipient_id: str = None,
        sender_public_key: bytes = None,
        requester_public_key: bytes = None,
        signature: bytes = None,
        timestamp: int = None,
        asset: LiskTransactionAsset = None,
    ) -> None:
        self.type = type
        self.amount = amount
        self.fee = fee
        self.recipient_id = recipient_id
        self.sender_public_key = sender_public_key
        self.requester_public_key = requester_public_key
        self.signature = signature
        self.timestamp = timestamp
        self.asset = asset



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class NEMProvisionNamespace(p.MessageType):
    FIELDS = {
        1: ('namespace', p.UnicodeType, 0),
        2: ('parent', p.UnicodeType, 0),
        3: ('sink', p.UnicodeType, 0),
        4: ('fee', p.UVarintType, 0),
    }

    def __init__(
        self,
        namespace: str = None,
        parent: str = None,
        sink: str = None,
        fee: int = None,
    ) -> None:
        self.namespace = namespace
        self.parent = parent
        self.sink = sink
        self.fee = fee



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p
if __debug__:
    try:
        from typing import List
    except ImportError:
        List = None  # type: ignore


class MoneroTransactionInitAck(p.MessageType):
    MESSAGE_WIRE_TYPE = 502
    FIELDS = {
        1: ('version', p.UVarintType, 0),
        2: ('status', p.UVarintType, 0),
        3: ('in_memory', p.BoolType, 0),
        4: ('hmacs', p.BytesType, p.FLAG_REPEATED),
        5: ('many_inputs', p.BoolType, 0),
        6: ('many_outputs', p.BoolType, 0),
    }

    def __init__(
        self,
        version: int = None,
        status: int = None,
        in_memory: bool = None,
        hmacs: List[bytes] = None,
        many_inputs: bool = None,
        many_outputs: bool = None,
    ) -> None:
        self.version = version
        self.status = status
        self.in_memory = in_memory
        self.hmacs = hmacs if hmacs is not None else []
        self.many_inputs = many_inputs
        self.many_outputs = many_outputs



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class StellarManageDataOp(p.MessageType):
    MESSAGE_WIRE_TYPE = 220
    FIELDS = {
        1: ('source_account', p.UnicodeType, 0),
        2: ('key', p.UnicodeType, 0),
        3: ('value', p.BytesType, 0),
    }

    def __init__(
        self,
        source_account: str = None,
        key: str = None,
        value: bytes = None,
    ) -> None:
        self.source_account = source_account
        self.key = key
        self.value = value



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p
if __debug__:
    try:
        from typing import List
    except ImportError:
        List = None  # type: ignore
from .TezosDelegationType import TezosDelegationType
from .TezosOperationCommon import TezosOperationCommon
from .TezosOriginationType import TezosOriginationType
from .TezosTransactionType import TezosTransactionType


class TezosSignTx(p.MessageType):
    MESSAGE_WIRE_TYPE = 152
    FIELDS = {
        1: ('address_n', p.UVarintType, p.FLAG_REPEATED),
        2: ('operation', TezosOperationCommon, 0),
        3: ('transaction', TezosTransactionType, 0),
        4: ('origination', TezosOriginationType, 0),
        5: ('delegation', TezosDelegationType, 0),
    }

    def __init__(
        self,
        address_n: List[int] = None,
        operation: TezosOperationCommon = None,
        transaction: TezosTransactionType = None,
        origination: TezosOriginationType = None,
        delegation: TezosDelegationType = None,
    ) -> None:
        self.address_n = address_n if address_n is not None else []
        self.operation = operation
        self.transaction = transaction
        self.origination = origination
        self.delegation = delegation



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p
from .MoneroRingCtSig import MoneroRingCtSig


class MoneroTransactionAllOutSetAck(p.MessageType):
    MESSAGE_WIRE_TYPE = 507
    FIELDS = {
        1: ('extra', p.BytesType, 0),
        2: ('tx_prefix_hash', p.BytesType, 0),
        3: ('rv', MoneroRingCtSig, 0),
    }

    def __init__(
        self,
        extra: bytes = None,
        tx_prefix_hash: bytes = None,
        rv: MoneroRingCtSig = None,
    ) -> None:
        self.extra = extra
        self.tx_prefix_hash = tx_prefix_hash
        self.rv = rv



from monero_glue.xmr import crypto
from monero_glue.xmr.sub.xmr_net import NetworkTypes, net_version


def addr_to_hash(addr):
    """
    Creates hashable address representation
    :param addr:
    :return:
    """
    return bytes(addr.m_spend_public_key + addr.m_view_public_key)


def encode_addr(version, spend_pub, view_pub):
    """
    Encodes public keys as versions
    :param version:
    :param spend_pub:
    :param view_pub:
    :return:
    """
    buf = spend_pub + view_pub
    return crypto.xmr_base58_addr_encode_check(ord(version), bytes(buf))


def decode_addr(addr):
    """
    Given address, get version and public spend and view keys.

    :param addr:
    :return:
    """
    d, version = crypto.xmr_base58_addr_decode_check(bytes(addr))
    pub_spend_key = d[0:32]
    pub_view_key = d[32:64]
    return version, pub_spend_key, pub_view_key


def public_addr_encode(pub_addr, is_sub=False, net=NetworkTypes.MAINNET):
    """
    Encodes public address to Monero address
    :param pub_addr:
    :type pub_addr: apps.monero.xmr.serialize_messages.addr.AccountPublicAddress
    :param is_sub:
    :param net:
    :return:
    """
    net_ver = net_version(net, is_sub)
    return encode_addr(net_ver, pub_addr.m_spend_public_key, pub_addr.m_view_public_key)


def classify_subaddresses(tx_dests, change_addr):
    """
    Classify destination subaddresses
    void classify_addresses()
    :param tx_dests:
    :type tx_dests: list[apps.monero.xmr.serialize_messages.tx_construct.TxDestinationEntry]
    :param change_addr:
    :return:
    """
    num_stdaddresses = 0
    num_subaddresses = 0
    single_dest_subaddress = None
    addr_set = set()
    for tx in tx_dests:
        if change_addr and addr_eq(change_addr, tx.addr):
            continue
        addr_hashed = addr_to_hash(tx.addr)
        if addr_hashed in addr_set:
            continue
        addr_set.add(addr_hashed)
        if tx.is_subaddress:
            num_subaddresses += 1
            single_dest_subaddress = tx.addr
        else:
            num_stdaddresses += 1
    return num_stdaddresses, num_subaddresses, single_dest_subaddress


def addr_eq(a, b):
    """
    Address comparisson. Allocation free.
    :param a:
    :param b:
    :return:
    """
    return pub_eq(a.m_spend_public_key, b.m_spend_public_key) and pub_eq(
        a.m_view_public_key, b.m_view_public_key
    )


def pub_eq(a, b):
    """
    Simple non-constant time public key compare
    :param a:
    :param b:
    :return:
    """
    if a == b:
        return True
    if (a is None and b is not None) or (a is not None and b is None):
        return False
    if len(a) != len(b):
        return False
    for i in range(len(a)):
        if a[i] != b[i]:
            return False
    return True


def get_change_addr_idx(outputs, change_dts):
    """
    Returns ID of the change output from the change_dts and outputs
    :param tsx_data:
    :return:
    """
    if change_dts is None:
        return None

    change_idx = None
    change_coord = change_dts.amount, change_dts.addr
    for idx, dst in enumerate(outputs):
        if (
            change_coord
            and change_coord[0]
            and change_coord[0] == dst.amount
            and addr_eq(change_coord[1], dst.addr)
        ):
            change_idx = idx
    return change_idx



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p
from .MoneroTransactionAllOutSetRequest import MoneroTransactionAllOutSetRequest
from .MoneroTransactionFinalRequest import MoneroTransactionFinalRequest
from .MoneroTransactionInitRequest import MoneroTransactionInitRequest
from .MoneroTransactionInputViniRequest import MoneroTransactionInputViniRequest
from .MoneroTransactionInputsPermutationRequest import MoneroTransactionInputsPermutationRequest
from .MoneroTransactionMlsagDoneRequest import MoneroTransactionMlsagDoneRequest
from .MoneroTransactionSetInputRequest import MoneroTransactionSetInputRequest
from .MoneroTransactionSetOutputRequest import MoneroTransactionSetOutputRequest
from .MoneroTransactionSignInputRequest import MoneroTransactionSignInputRequest


class MoneroTransactionSignRequest(p.MessageType):
    MESSAGE_WIRE_TYPE = 501
    FIELDS = {
        1: ('init', MoneroTransactionInitRequest, 0),
        2: ('set_input', MoneroTransactionSetInputRequest, 0),
        3: ('input_permutation', MoneroTransactionInputsPermutationRequest, 0),
        4: ('input_vini', MoneroTransactionInputViniRequest, 0),
        5: ('set_output', MoneroTransactionSetOutputRequest, 0),
        6: ('all_out_set', MoneroTransactionAllOutSetRequest, 0),
        7: ('mlsag_done', MoneroTransactionMlsagDoneRequest, 0),
        8: ('sign_input', MoneroTransactionSignInputRequest, 0),
        9: ('final_msg', MoneroTransactionFinalRequest, 0),
    }

    def __init__(
        self,
        init: MoneroTransactionInitRequest = None,
        set_input: MoneroTransactionSetInputRequest = None,
        input_permutation: MoneroTransactionInputsPermutationRequest = None,
        input_vini: MoneroTransactionInputViniRequest = None,
        set_output: MoneroTransactionSetOutputRequest = None,
        all_out_set: MoneroTransactionAllOutSetRequest = None,
        mlsag_done: MoneroTransactionMlsagDoneRequest = None,
        sign_input: MoneroTransactionSignInputRequest = None,
        final_msg: MoneroTransactionFinalRequest = None,
    ) -> None:
        self.init = init
        self.set_input = set_input
        self.input_permutation = input_permutation
        self.input_vini = input_vini
        self.set_output = set_output
        self.all_out_set = all_out_set
        self.mlsag_done = mlsag_done
        self.sign_input = sign_input
        self.final_msg = final_msg



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p
if __debug__:
    try:
        from typing import List
    except ImportError:
        List = None  # type: ignore
from .HDNodePathType import HDNodePathType


class MultisigRedeemScriptType(p.MessageType):
    FIELDS = {
        1: ('pubkeys', HDNodePathType, p.FLAG_REPEATED),
        2: ('signatures', p.BytesType, p.FLAG_REPEATED),
        3: ('m', p.UVarintType, 0),
    }

    def __init__(
        self,
        pubkeys: List[HDNodePathType] = None,
        signatures: List[bytes] = None,
        m: int = None,
    ) -> None:
        self.pubkeys = pubkeys if pubkeys is not None else []
        self.signatures = signatures if signatures is not None else []
        self.m = m



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class TezosContractID(p.MessageType):
    FIELDS = {
        1: ('tag', p.UVarintType, 0),
        2: ('hash', p.BytesType, 0),
    }

    def __init__(
        self,
        tag: int = None,
        hash: bytes = None,
    ) -> None:
        self.tag = tag
        self.hash = hash



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class CipheredKeyValue(p.MessageType):
    MESSAGE_WIRE_TYPE = 48
    FIELDS = {
        1: ('value', p.BytesType, 0),
    }

    def __init__(
        self,
        value: bytes = None,
    ) -> None:
        self.value = value



from monero_serialize import xmrserialize
from monero_glue.xmr import crypto


class KeccakArchive(object):
    def __init__(self, ctx=None):
        self.kwriter = get_keccak_writer(ctx=ctx)
        self.ar = xmrserialize.Archive(self.kwriter, True)

    def ctx(self):
        return self.kwriter.ctx()

    def refresh(self, ctx=None, xser=None):
        if ctx is None:
            ctx = self.kwriter.ctx()
        if xser is None:
            xser = xmrserialize

        self.kwriter = get_keccak_writer(ctx=ctx)
        self.ar = xser.Archive(self.kwriter, True)
        return self.ar


class HashWrapper(object):
    def __init__(self, ctx):
        self.ctx = ctx

    def update(self, buf):
        if len(buf) == 0:
            return
        self.ctx.update(buf)

    def digest(self):
        return self.ctx.digest()

    def hexdigest(self):
        return self.ctx.hexdigest()


class AHashWriter:
    def __init__(self, hasher, sub_writer=None):
        self.hasher = hasher
        self.sub_writer = sub_writer

    async def awrite(self, buf):
        self.hasher.update(buf)
        if self.sub_writer:
            await self.sub_writer.awrite(buf)
        return len(buf)

    def get_digest(self, *args) -> bytes:
        return self.hasher.digest(*args)

    def ctx(self):
        return self.hasher.ctx


def get_keccak_writer(sub_writer=None, ctx=None):
    """
    Creates new fresh async Keccak writer
    :param sub_writer:
    :param ctx:
    :return:
    """
    return AHashWriter(
        HashWrapper(crypto.get_keccak() if ctx is None else ctx), sub_writer=sub_writer
    )



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p
if __debug__:
    try:
        from typing import List
    except ImportError:
        List = None  # type: ignore
from .RipplePayment import RipplePayment


class RippleSignTx(p.MessageType):
    MESSAGE_WIRE_TYPE = 402
    FIELDS = {
        1: ('address_n', p.UVarintType, p.FLAG_REPEATED),
        2: ('fee', p.UVarintType, 0),
        3: ('flags', p.UVarintType, 0),
        4: ('sequence', p.UVarintType, 0),
        5: ('last_ledger_sequence', p.UVarintType, 0),
        6: ('payment', RipplePayment, 0),
    }

    def __init__(
        self,
        address_n: List[int] = None,
        fee: int = None,
        flags: int = None,
        sequence: int = None,
        last_ledger_sequence: int = None,
        payment: RipplePayment = None,
    ) -> None:
        self.address_n = address_n if address_n is not None else []
        self.fee = fee
        self.flags = flags
        self.sequence = sequence
        self.last_ledger_sequence = last_ledger_sequence
        self.payment = payment



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class MoneroTransactionSetInputRequest(p.MessageType):
    FIELDS = {
        1: ('version', p.UVarintType, 0),
        2: ('src_entr', p.BytesType, 0),
    }

    def __init__(
        self,
        version: int = None,
        src_entr: bytes = None,
    ) -> None:
        self.version = version
        self.src_entr = src_entr



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p
from .NEMMosaicDefinition import NEMMosaicDefinition


class NEMMosaicCreation(p.MessageType):
    FIELDS = {
        1: ('definition', NEMMosaicDefinition, 0),
        2: ('sink', p.UnicodeType, 0),
        3: ('fee', p.UVarintType, 0),
    }

    def __init__(
        self,
        definition: NEMMosaicDefinition = None,
        sink: str = None,
        fee: int = None,
    ) -> None:
        self.definition = definition
        self.sink = sink
        self.fee = fee
